@@ -1,7 +1,7 @@
 #!/usr/bin/env python3
 # vim: set expandtab tabstop=4 shiftwidth=4:
 
-# Copyright (c) 2020-2021 CJ Kucera (cj@apocalyptech.com)
+# Copyright (c) 2020-2022 CJ Kucera (cj@apocalyptech.com)
 # 
 # This software is provided 'as-is', without any express or implied warranty.
 # In no event will the authors be held liable for any damages arising from
@@ -31,6 +31,7 @@ import binascii
 import pkg_resources
 
 from . import *
+from . import OakShared_pb2
 
 class ArbitraryBits(object):
     """
@@ -101,12 +102,26 @@ class ArbitraryBits(object):
             byte_data.append(int(temp_data[i*8:(i*8)+8], 2))
         return bytearray(byte_data)
 
-class BL3Serial(object):
+class WLSerial(object):
     """
-    Class to handle serializing and deserializing BL3 item/weapon serial
+    Class to handle serializing and deserializing WL item/weapon serial
     numbers.
     """
 
+    # Supported item-code prefixes for importing
+    code_prefixes = {
+            # New standard "we" settled on, 2022-07-17
+            'wl',
+            # The original prefix Arwent used when releasing the first editor;
+            # the de facto standard for a few months.
+            'ttw',
+            # Not actually widely in-use, but this repo technically used it for
+            # awhile, thinking that it was the de facto standard (it was not).
+            'ttwl',
+            # too much code still uses bl3 -- deprecate me
+            'bl3',
+            }
+
     def __init__(self, serial, datawrapper):
 
         self.datawrapper = datawrapper
@@ -129,7 +144,7 @@ class BL3Serial(object):
         """
 
         self.serial = serial
-        (self.decrypted_serial, self.orig_seed, self.serial_version) = BL3Serial._decrypt_serial(serial)
+        (self.decrypted_serial, self.orig_seed, self.serial_version) = WLSerial._decrypt_serial(serial)
         self.parsed = False
         self.parts_parsed = False
         self.can_parse = True
@@ -161,7 +176,7 @@ class BL3Serial(object):
         self._generic_parts = None
         self._additional_data = None
         self._num_customs = None
-        self._item_type = None
+        self._chaos_level = None
         
         # Call out to any superclass procedures here
         self._update_superclass_serial()
@@ -195,7 +210,7 @@ class BL3Serial(object):
         """
 
         # First run it through the xor wringer.
-        temp = BL3Serial._xor_data(data, seed)
+        temp = WLSerial._xor_data(data, seed)
 
         # Now rotate the data
         steps = (seed & 0x1F) % len(data)
@@ -212,7 +227,7 @@ class BL3Serial(object):
         rotated = bytearray(data[steps:] + data[:steps])
 
         # Then run through the xor stuff
-        return bytearray(BL3Serial._xor_data(rotated, seed))
+        return bytearray(WLSerial._xor_data(rotated, seed))
 
     @staticmethod
     def _decrypt_serial(serial):
@@ -227,14 +242,14 @@ class BL3Serial(object):
             #print(dir(serial))
             #print(f'ss:{serial.item_serial_number}')
             serial = serial.item_serial_number
-        assert(serial[0] in [3,4,5])
+        assert(serial[0] in {5})
         serial_version = serial[0]
 
         # Seed does need to be an unsigned int
         orig_seed = struct.unpack('>i', serial[1:5])[0]
 
         # Do the actual "decryption"
-        decrypted = BL3Serial._bogodecrypt(serial[5:], orig_seed)
+        decrypted = WLSerial._bogodecrypt(serial[5:], orig_seed)
 
         # Grab the CRC stored in the serial itself
         orig_checksum = bytearray(decrypted[:2])
@@ -273,7 +288,7 @@ class BL3Serial(object):
         checksum = struct.pack('>H', ((crc32 >> 16) ^ crc32) & 0xFFFF)
 
         # Return the freshly-encrypted item
-        return header + BL3Serial._bogoencrypt(checksum + data, seed)
+        return header + WLSerial._bogoencrypt(checksum + data, seed)
 
     def _get_inv_db_header_part(self, category, bits):
         """
@@ -377,7 +392,7 @@ class BL3Serial(object):
                     self._part_invkey, bits, 6)
             # print((self._part_bits, self._parts))
             # print(len(bits.data))
-            # Read generics (anointments+mayhem)
+            # Read generics (enchantments)
             (self._generic_bits, self._generic_parts) = self._get_inv_db_header_part_repeated(
                     'InventoryGenericPartData', bits, 4)
 
@@ -398,37 +413,34 @@ class BL3Serial(object):
                 self.parts_parsed = False
                 self.can_parse_parts = False
 
-            # If we're a v4 (or higher) serial, read in the number of times we've
-            # been re-rolled
+            # Read in the number of times we've been re-rolled
             if (len(bits.data) >= 8):
-                if self.serial_version >= 4:
-                    self._rerolled = bits.eat(8)
-                else:
-                    self._rerolled = 0
-            # AH: Not sure if we need this
-            # # And read in our remaining data.  If there's more than 7 bits
-            # # left, we've done something wrong, because it should only be
-            # # zero-padding after all the "real" data is in place.
-            # if len(bits.data) > 7:
-            #     self.parts_parsed = False
-            #     self.can_parse_parts = False
-            #     pass
-            # elif '1' in bits.data:
-            #     # This is supposed to only be zero-padding at the moment, if
-            #     # we see something else, abort
-            #     self.parts_parsed = False
-            #     self.can_parse_parts = False
-            # else:
-            #     # Okay, we're good!  Don't bother saving the remaining 0 bits.
-            #     pass
-            if len(bits.data) > 3:                
-                self._item_type = bits.eat(3)
+                self._rerolled = bits.eat(8)
+
+            # Enhancement / Chaos Level (? - not sure what to call this)
+            self._chaos_level = bits.eat(7)
+
+            # And read in our remaining data.  If there's more than 7 bits
+            # left, we've done something wrong, because it should only be
+            # zero-padding after all the "real" data is in place.
+            if len(bits.data) > 7:
+                self.parts_parsed = False
+                self.can_parse_parts = False
+                pass
+            elif '1' in bits.data:
+                # This is supposed to only be zero-padding at the moment, if
+                # we see something else, abort
+                self.parts_parsed = False
+                self.can_parse_parts = False
+            else:
+                # Okay, we're good!  Don't bother saving the remaining 0 bits.
+                pass
 
     def _deparse_serial(self):
         """
         De-parses a serial; used after we make changes to the data that gets
         pulled out during `_parse_serial`.  At the moment, that's both level
-        changes and mayhem level changes.  Will end up calling out to the
+        changes and chaos level changes.  Will end up calling out to the
         superclass's `_update_superclass_serial` to propagate the serial change
         to whatever containing structure needs it, and set the object to trigger a
         re-parse if anything else needs to read more.  That's probably overkill
@@ -485,12 +497,11 @@ class BL3Serial(object):
             # Then our number of customs (should always be zero)
             bits.append_value(self._num_customs, 4)
 
-            # Then, if we're a v4 serial, the number of times we've been rerolled
-            if self.serial_version >= 4:
-                bits.append_value(self._rerolled, 8)
-            # Then, if we're a v4 serial, the item_type
-            if self.serial_version >= 4:
-                bits.append_value(self._item_type, 3)
+            # Then the number of times we've been rerolled
+            bits.append_value(self._rerolled, 8)
+
+            # Then the chaos evel
+            bits.append_value(self._chaos_level, 7)
 
         else:
             # Otherwise, we can re-use our original remaining data
@@ -500,7 +511,7 @@ class BL3Serial(object):
         new_data = bits.get_data()
 
         # Encode the new serial (using seed 0; unencrypted)
-        new_serial = BL3Serial._encrypt_serial(new_data, self.serial_version, 0)
+        new_serial = WLSerial._encrypt_serial(new_data, self.serial_version, 0)
 
         # Load in the new serial (this will set `parsed` to `False`)
         # It bothers me that I've just done an `_encrypt_serial` in the
@@ -585,7 +596,38 @@ class BL3Serial(object):
             seed = self.orig_seed
         else:
             seed = 0
-        return BL3Serial._encrypt_serial(self.decrypted_serial, self.serial_version, seed)
+        return WLSerial._encrypt_serial(self.decrypted_serial, self.serial_version, seed)
+
+    @property
+    def rerolled(self):
+        """
+        Returns the number of times this item's enchantments have been rerolled
+        """
+        if not self.parsed:
+            self._parse_serial()
+            if not self.can_parse:
+                return None
+        return self._rerolled
+
+    @rerolled.setter
+    def rerolled(self, value):
+        """
+        Sets a new reroll count for the item.  This is a super inefficient way of
+        doing it -- we're rebuilding the whole serial right now and triggering a re-parse
+        if anything decides to re-read it.  It's, y'know, *sufficient*, but ick.
+        """
+        if not self.parsed:
+            self._parse_serial()
+            if not self.can_parse:
+                return None
+
+        # Set the reroll count and trigger a re-encode of the serial
+        self._rerolled = value
+        # Eh, we need this 'cause it's stored *after* parts.  I really do need to
+        # make this whole item-editing API about a million times more sensible.
+        self.changed_parts = True
+        self._deparse_serial()
+        self._update_superclass_serial()
 
     def get_serial_base64(self, orig_seed=False):
         """
@@ -594,194 +636,138 @@ class BL3Serial(object):
         savegame.  Otherwise, it will use a seed of `0`, which will then be
         unencrypted.
         """
-        return 'BL3({})'.format(base64.b64encode(self.get_serial_number(orig_seed)).decode('latin1'))
+        return 'WL({})'.format(base64.b64encode(self.get_serial_number(orig_seed)).decode('latin1'))
+
+    @staticmethod
+    def get_inner_serial_base64(serial):
+        """
+        Given a `WL()`-encoded serial, return the actual inner code without the
+        syntactic sugar around it.  Checks for a variety of valid prefixes,
+        rather than just `WL`.  Returns `None` if there was no valid prefix
+        found.
+        """
+        parts = serial.split('(')
+        if len(parts) != 2:
+            return None
+        if parts[0].lower() not in WLSerial.code_prefixes:
+            return None
+        if parts[1][-1] != ')':
+            return None
+        return parts[1][:-1]
 
     @staticmethod
     def decode_serial_base64(new_data):
         """
-        Decodes a `BL3()`-encoded item serial into a binary serial
+        Decodes a `WL()`-encoded item serial into a binary serial
         """
-        if not new_data.lower().startswith('bl3(') or not new_data.endswith(')'):
+        inner_serial = WLSerial.get_inner_serial_base64(new_data)
+        if inner_serial:
+            return base64.b64decode(inner_serial)
+        else:
             raise Exception('Unknown item format: {}'.format(new_data))
-        encoded = new_data[4:-1]
-        return base64.b64decode(encoded)
+
+    def can_have_enchantment(self):
+        """
+        Returns `True` if this is an item type which can have an enchantment,
+        or `False` otherwise.  Will also return `False` if we're unable to
+        parse parts for the item.
+        """
+        if not self.parsed or not self.parts_parsed:
+            self._parse_serial()
+            if not self.can_parse or not self.can_parse_parts:
+                return False
+        return self._invdata.lower() in enchantment_invdata_lower_types
+
+    def can_have_chaos_level(self):
+        """
+        Returns `True` if this is an item type which can have a chaos level,
+        or `False` otherwise.  Will also return `False` if we're unable to
+        parse parts for the item.
+        """
+        if not self.parsed or not self.parts_parsed:
+            self._parse_serial()
+            if not self.can_parse or not self.can_parse_parts:
+                return False
+        return self._chaos_level is not None
 
     @property
-    def mayhem_level(self):
+    def chaos_level(self):
         """
-        Returns the current Mayhem level of the item, with `0` signifying
-        that there is no Mayhem level present, and `None` signifying that
-        the Mayhem level could not be parsed (due to being unable to parse the
-        item parts)
+        Returns the current chaos level (0 is basic, 1 is chaotic, etc),
+        or `None` signifying that the serial could not be parsed
         """
         if not self.parsed or not self.parts_parsed:
             self._parse_serial()
             if not self.can_parse or not self.can_parse_parts:
                 return None
-        # Given the presence of item editors, there could possibly be more
-        # than one Mayhem part present in a serial (though they don't seem
-        # to stack at all, so doing so would be pointless).  We'll just
-        # abort processing as soon as we find one, which I suspect is likely
-        # what the game does, too.
-        for part_name, part_idx in self._generic_parts:
-            if part_name.lower() in mayhem_part_lower_to_lvl:
-                return mayhem_part_lower_to_lvl[part_name.lower()]
-        return 0
-
-    def can_have_mayhem(self):
-        """
-        Returns `True` if this is an item type which can have a mayhem level,
-        or `False` otherwise.  Will also return `False` if we're unable to
-        parse parts for the item.
-        """
-        if not self.parsed or not self.parts_parsed:
-            self._parse_serial()
-            if not self.can_parse or not self.can_parse_parts:
-                return False
-        return self._invdata.lower() in mayhem_invdata_lower_types
-
-    def can_have_anointment(self):
-        """
-        Returns `True` if this is an item type which can have an anointment,
-        or `False` otherwise.  Will also return `False` if we're unable to
-        parse parts for the item.
-        """
-        if not self.parsed or not self.parts_parsed:
-            self._parse_serial()
-            if not self.can_parse or not self.can_parse_parts:
-                return False
-        return self._invdata.lower() in anointable_invdata_lower_types
-
-    @mayhem_level.setter
-    def mayhem_level(self, value):
-        """
-        Sets the given mayhem level on the item.  Returns `True` if we were
-        able to do so, or `False` if not.
-        """
-        # The call to `can_have_mayhem` will parse the serial if possible,
-        # so we'll be all set.
-        if not self.can_have_mayhem():
-            return False
-
-        # Don't forget to set this
-        self.changed_parts = True
-
-        # First grab a list of any non-Mayhem parts (should just be anoints)
-        new_parts = []
-        for idx, (part_name, part_idx) in enumerate(self._generic_parts):
-            if part_name.lower() not in mayhem_part_lower_to_lvl:
-                new_parts.append((part_name, part_idx))
-
-        # Now add our new one in
-        if value > 0:
-            new_mayhem_part = self.serial_db.get_part_index(
-                    'InventoryGenericPartData',
-                    mayhem_lvl_to_part[value],
-                    )
-            if new_mayhem_part is None:
-                return False
-            else:
-                new_parts.append((mayhem_lvl_to_part[value], new_mayhem_part))
-
-        # Aaaand assign our list of generic parts back
-        self._generic_parts = new_parts
-
-        # Re-serialize
-        self._deparse_serial()
-        self._update_superclass_serial()
-
-        # return!
-        return True
-
-    def can_have_item_type(self):
-        """
-        Returns `True` if this is an item type which can have a mayhem level,
-        or `False` otherwise.  Will also return `False` if we're unable to
-        parse parts for the item.
-        """
-        if not self.parsed or not self.parts_parsed:
-            self._parse_serial()
-            if not self.can_parse or not self.can_parse_parts:
-                return False
-        return self._item_type is not None
-
-    @property
-    def item_type(self):
-        """
-        Returns the current item_type 0 being basic, 1 being chaotic. 
-         and `None` signifying that
-        the item_type level could not be parsed
-        """
-        if not self.parsed or not self.parts_parsed:
-            self._parse_serial()
-            if not self.can_parse or not self.can_parse_parts:
-                return None
-        return self._item_type
+        return self._chaos_level
 
     
-    @item_type.setter
-    def item_type(self, value):
+    @chaos_level.setter
+    def chaos_level(self, value):
         """
-        Sets the given item_type level on the item.  Returns `True` if we were
+        Sets the given chaos level level on the item.  Returns `True` if we were
         able to do so, or `False` if not.
         """
-        # The call to `can_have_mayhem` will parse the serial if possible,
+        # The call to `can_have_chaos_level` will parse the serial if possible,
         # so we'll be all set.
-        if not self.can_have_item_type():
+        if not self.can_have_chaos_level():
             return False
         # Don't forget to set this
         self.changed_parts = True
-        self._item_type = value
+        self._chaos_level = value
         # Re-serialize
         self._deparse_serial()
         self._update_superclass_serial()
         # return!
         return True
 
+
+    def get_chaos_level_eng(self):
+        level = ChaosLevel.has_value(self.chaos_level)
+        if level:
+            return level.label
+        else:
+            return None
     
-    def set_anointment(self, anointment):
+    def set_enchantment(self, enchantment):
         """
-        Sets the given anointment on the item, if possible.  Returns `True` if we were
+        Sets the given enchantment on the item, if possible.  Returns `True` if we were
         able to do so, or `False` if not.  This does not do any checking to see if
-        the anointment would be ordinarily "valid" for the given item type, but it does
+        the enchantment would be ordinarily "valid" for the given item type, but it does
         at least attempt to only apply if it's an item type which can ordinarily have
-        anointments.
+        enchantments.
 
-        This will overwrite any existing anointment part on the item in question.
+        This will overwrite any existing enchantment part on the item in question.
 
-        TODO: This routine currently assumes that any Generic part that's *not* a
-        Mayhem Level part is an anointment, and will wipe those out before setting
-        the specified part.  As of November 2020 this should be a safe assumption,
-        but might not be in the future.  Should this functionality ever get exported
-        though a "proper" CLI arg, we should really import a list of legit
-        anointments to check against, just to assist in futureproofing.
+        TODO: This routine currently assumes that any existing Generic part is an
+        enchantment, and will wipe those out before setting the specified part.
+        As of July 2022 this should be a safe assumption, but might not be in
+        the future.  Should this functionality ever get exported though a
+        "proper" CLI arg, we should really import a list of legit enchantments
+        to check against, just to assist in futureproofing.
         """
+        # TODO: Not actually tested *at all* in Wonderlands yet.
 
-        # Check for anointment part validity first thing, before we do anything
+        # Check for enchantment part validity first thing, before we do anything
         # else.
-        new_anointment_part = self.serial_db.get_part_index(
+        new_enchantment_part = self.serial_db.get_part_index(
                 'InventoryGenericPartData',
-                anointment,
+                enchantment,
                 )
-        if not new_anointment_part:
-            raise Exception('ERROR: {} is not a known anointment'.format(anointment))
+        if not new_enchantment_part:
+            raise Exception('ERROR: {} is not a known enchantment'.format(enchantment))
 
-        # The call to `can_have_anointment` will parse the serial if possible,
+        # The call to `can_have_enchantment` will parse the serial if possible,
         # so we'll be all set.
-        if not self.can_have_anointment():
+        if not self.can_have_enchantment():
             return False
 
         # Don't forget to set this
         self.changed_parts = True
 
-        # Start out with our new anointment part
-        new_parts = [(anointment, new_anointment_part)]
-
-        # Now add in any existing Mayhem parts (should just be the one, but
-        # whatever)
-        for idx, (part_name, part_idx) in enumerate(self._generic_parts):
-            if part_name.lower() in mayhem_part_lower_to_lvl:
-                new_parts.append((part_name, part_idx))
+        # Start out with our new enchantment part
+        new_parts = [(enchantment, new_enchantment_part)]
 
         # Aaaand assign our list of generic parts back
         self._generic_parts = new_parts
@@ -792,13 +778,10 @@ class BL3Serial(object):
 
         # return!
         return True
-    def get_item_type_eng(self):
-        it = self.item_type
-        return item_type_eng.get(it,it)
         
     def get_level_eng(self):
         """
-        Returns an English representation of our level, including Mayhem level,
+        Returns an English representation of our level, including Chaos level,
         suitable for reporting to a user.
         """
         # First, regular level
@@ -807,15 +790,67 @@ class BL3Serial(object):
             return 'unknown lvl'
         to_ret = 'level {}'.format(level)
 
-        # Then Mayhem
-        mayhem_level = self.mayhem_level
-        if mayhem_level is None:
-            return '{}, mayhem unknown'.format(to_ret)
-        elif mayhem_level > 0:
-            return '{}, mayhem {}'.format(to_ret, mayhem_level)
+        # Then Chaos Level
+        chaos_label = self.get_chaos_level_eng()
+        if chaos_label is None:
+            return '{}, chaos unknown'.format(to_ret)
+        elif self.chaos_level > 0:
+            return '{}, {}'.format(to_ret, chaos_label)
         else:
             return to_ret
 
+class WLItem(WLSerial):
+    """
+    Pretty thin wrapper around the protobuf `OakInventoryItemSaveGameData`
+    object for an item, used by both the player inventory and the bank.
+    (Though not the Lost Loot machine!  That is just an array of raw
+    serials.)  We're ignoring `development_save_data` entirely since it
+    doesn't seem to be present in actual savegames.
+
+    No idea what `pickup_order_index` is, though it might just have
+    something to do with the ordering when you're picking up multiple
+    things at once (in which case it's probably only really useful for
+    things like money and ammo).
+    """
+
+    def __init__(self, protobuf, datawrapper):
+        self.protobuf = protobuf
+        super().__init__(self.protobuf.item_serial_number, datawrapper)
+
+    @staticmethod
+    def create(datawrapper, serial_number, pickup_order_idx, is_seen=True, is_favorite=False, is_trash=False):
+        """
+        Creates a new item with the specified serial number, and pickup_order_idx
+        """
+
+        # Start constructing flags
+        flags = 0
+        if is_seen:
+            flags |= 0x1
+
+        # Favorite and Trash are mutually-exclusive
+        if is_favorite:
+            flags |= 0x2
+        elif is_trash:
+            flags |= 0x4
+
+        # Now do the creation
+        return WLItem(OakShared_pb2.OakInventoryItemSaveGameData(
+                item_serial_number=serial_number,
+                pickup_order_index=pickup_order_idx,
+                flags=flags,
+                ), datawrapper)
+
+    def get_pickup_order_idx(self):
+        return self.protobuf.pickup_order_index
+
+    def _update_superclass_serial(self):
+        """
+        Action to take when our serial number gets updated.  In this case,
+        setting the serial back into the protobuf.
+        """
+        self.protobuf.item_serial_number = self.serial
+
 class InventorySerialDB(object):
     """
     Little wrapper to provide access to our inventory serial number DB
@@ -978,7 +1013,7 @@ class DataWrapper(object):
     Weird little metaclass which just has an instance of each of our file-backed
     data objects in here.  This way apps using it can just pass around a single
     object instance and take what they want, rather than having to carry around
-    multiple.  (For instance, BL3Item needs both InventorySerialDB and
+    multiple.  (For instance, WLItem needs both InventorySerialDB and
     BalanceToName, and we instantiate a fair number of those.)
     """
 
